/* This code is provided under the terms of the 
 * MIT License, see at the end of the file
 * (c) 2013 Alejandro Paz
 */



`timescale 1ns/100ps
`default_nettype none


`define OP_ARITHM_ADDBCD  5'b00000
`define OP_ARITHM_ADDBIN  5'b00001
`define OP_ARITHM_AND     5'b00010
`define OP_ARITHM_OR      5'b00011
`define OP_ARITHM_SUBBCD  5'b00100
`define OP_ARITHM_SUBBIN  5'b00101
`define OP_ARITHM_NOT     5'b00110
`define OP_ARITHM_COMP9   5'b00111

`define OP_ARITHM_ADCBCD  5'b01000
`define OP_ARITHM_ADCBIN  5'b01001
`define OP_ARITHM_NAND    5'b01010
`define OP_ARITHM_XOR     5'b01011
`define OP_ARITHM_SBCBCD  5'b01100
`define OP_ARITHM_SBCBIN  5'b01101
`define OP_ARITHM_MOV     5'b01110
`define OP_ARITHM_MUL     5'b01111

`define OP_ARITHM_SHL     5'b10000
`define OP_ARITHM_SHR     5'b10001
`define OP_ARITHM_ROL     5'b10010
`define OP_ARITHM_ROR     5'b10011
`define OP_ARITHM_SHL4    5'b10100
`define OP_ARITHM_SHR4    5'b10101
`define OP_ARITHM_ROL4    5'b10110
`define OP_ARITHM_ROR4    5'b10111

`define OP_JUMP_JMP       5'b11000
`define OP_JUMP_DJNZ      5'b11001
`define OP_JUMP_TJNZ      5'b11010
`define OP_JUMP_TJZ       5'b11011
`define OP_XTRA_RDBYTE    5'b11100
`define OP_XTRA_WRBYTE    5'b11101
`define OP_XTRA_IN        5'b11110
`define OP_XTRA_OUT       5'b11111

/**
 *
 * Instruction encoding
 *          5                   3              3          1                  10                                         10
 * +---+---+---+---+---+ +---+---+---+ +----+----+----+ +---+ +---+---+---+---+---+---+---+---+---+  +---+---+---+---+---+---+---+---+---+---+
 * | O   O   O   O   O | | C   C   C | | wz | wc | wr | | i | |                                   |  |             S   R   E   G             |
 * +---+---+---+---+---+ +---+---+---+ +----+----+----+ +---+ +---+---+---+---+---+---+---+---+---+  +---+---+---+---+---+---+---+---+---+---+
 * 
 * 
 * 
 * 
 * 
 */
`define BIT_I 20
`define FIELD_S 9:0

module tb_v8();
`define RDELAY #5
reg clk, reset;
wire [31:0] output_reg;

pnut_v8_core core(
    .CLK_I(clk),
    .RESET_I(reset),
    .PORT_O(output_reg)
    );

always
    #50 clk =~clk;

initial
  begin
    $dumpvars;
    #0
    reset = 1;
    clk = 0;
    #5000
    $finish;
  end
  
endmodule

module pnut_v8_core(
    CLK_I,
    RESET_I,
    PORT_O
    );
    
input CLK_I, RESET_I;
output [31:0] PORT_O;

wire internal_clk, internal_reset;

wire fetch_opcode_t1;
wire fetch_sd_t1;
wire tfr_sd_t1;
wire wb_d_with_alu_t1;
wire wb_d_with_pcp1_t1;
wire wb_flags_t1;
wire wb_pc_with_pcp1_t1;
wire wb_pc_with_s_t1;

wire fetch_opcode_t2;
wire fetch_sd_t2;
wire tfr_sd_t2;
wire wb_d_with_alu_t2;
wire wb_d_with_pcp1_t2;
wire wb_flags_t2;
wire wb_pc_with_pcp1_t2;
wire wb_pc_with_s_t2;

wire [31:0] opcode_t1, opcode_t2, preopcode_t1, preopcode_t2;
wire [31:0] data_d, data_s, alu_out, d_in;

reg [9:0] reg_pc_t1, reg_pc_t2;
reg reg_c_t1, reg_c_t2, reg_z_t1, reg_z_t2;

wire [9:0] pc_plus_1_t1, pc_plus_1_t2;

assign pc_plus_1_t1 = reg_pc_t1 + 9'h001;
assign pc_plus_1_t2 = reg_pc_t2 + 9'h001;

assign internal_clk = CLK_I;
assign internal_reset = RESET_I;

assign d_in = wb_d_with_pcp1_t1 ? { 22'b11000_111_000_0_0000000000, data_s[9:0] }:
              wb_d_with_pcp1_t2 ? { 22'b11000_111_000_0_0000000000, data_s[9:0] }:data_d;

pnut_v8_seqencer seq(
    .CLK_I(internal_clk),
    .RESET_I(internal_reset),
    .OPCODE_T1_I(opcode_t1[31:20]),
    .OPCODE_T2_I(opcode_t2[31:20]),
    .FLAG_Z_T1_I(reg_z_t1),
    .FLAG_C_T1_I(reg_c_t1),
    .FLAG_Z_T2_I(reg_z_t2),
    .FLAG_C_T2_I(reg_c_t2),
    /* T 1 */
    .FETCH_OPCODE_T1_O(fetch_opcode_t1),
    .FETCH_SD_T1_O(fetch_sd_t1),
    .TFR_SD_T1_O(tfr_sd_t1),
    .WRITEBACK_D_WITH_ALU_T1_O(wb_d_with_alu_t1),
    .WRITEBACK_D_WITH_PCP1_T1_O(wb_d_with_pcp1_t1),
    .WRITEBACK_FLAGS_T1_O(wb_flags_t1),
    .WRITEBACK_PC_WITH_PCP1_T1_O(wb_pc_with_pcp1_t1),
    .WRITEBACK_PC_WITH_S_T1_O(wb_pc_with_s_t1),
    
    /* T 2 */
    .FETCH_OPCODE_T2_O(fetch_opcode_t2),
    .FETCH_SD_T2_O(fetch_sd_t2),
    .TFR_SD_T2_O(tfr_sd_t2),
    .WRITEBACK_D_WITH_ALU_T2_O(wb_d_with_alu_t2),
    .WRITEBACK_D_WITH_PCP1_T2_O(wb_d_with_pcp1_t2),
    .WRITEBACK_FLAGS_T2_O(wb_flags_t2),
    .WRITEBACK_PC_WITH_PCP1_T2_O(wb_pc_with_pcp1_t2),
    .WRITEBACK_PC_WITH_S_T2_O(wb_pc_with_s_t2)
    );

pnut_v8_mem imem(
    .CLK_I(internal_clk),
	.RESET_I(internal_reset),
	
	.FETCH_OP_T1_I(fetch_opcode_t1),
	.FETCH_OP_T2_I(fetch_opcode_t2),
	.LOAD_OP_T1_I(fetch_sd_t1),
	.LOAD_OP_T2_I(fetch_sd_t2),
	.FETCH_SD_I(fetch_sd_t1 | fetch_sd_t2),
	.TFR_SD_I(tfr_sd_t1 | tfr_sd_t2),
	.WB_D_T1_I(wb_d_with_alu_t1 | wb_d_with_pcp1_t1),
	.WB_D_T2_I(wb_d_with_alu_t2 | wb_d_with_pcp1_t2),
	
	.D_I(d_in),
	.PC_T1_I(reg_pc_t1),
	.PC_T2_I(reg_pc_t2),
	.OPCODE_T1_O(opcode_t1),
	.OPCODE_T2_O(opcode_t2),
	.PREOPCODE_T1_O(preopcode_t1),
	.PREOPCODE_T2_O(preopcode_t2),
	
	.D_O(data_d),
	.S_O(data_s)
	);

always @(posedge internal_clk)
    begin
        if (wb_pc_with_pcp1_t1 == 1'b1)
            reg_pc_t1 <= pc_plus_1_t1;
        if (wb_pc_with_s_t1 == 1'b1)
            reg_pc_t1 <= data_s[9:0];
        
        if (wb_pc_with_pcp1_t2 == 1'b1)
            reg_pc_t2 <= pc_plus_1_t2;
        if (wb_pc_with_s_t2 == 1'b1)
            reg_pc_t2 <= data_s[9:0];
    end    
	
	
initial
    begin
        reg_pc_t1 = 0;
        reg_pc_t2 = 32;
        reg_c_t1 = 0;
        reg_c_t2 = 0; 
        reg_z_t1 = 0;
        reg_z_t2 = 0;
    end
endmodule


module pnut_v8_seqencer(
    CLK_I,
    RESET_I,
    OPCODE_T1_I,
    OPCODE_T2_I,
    FLAG_Z_T1_I,
    FLAG_C_T1_I,
    FLAG_Z_T2_I,
    FLAG_C_T2_I,
    /* T 1 */
    FETCH_OPCODE_T1_O,
    FETCH_SD_T1_O,
    TFR_SD_T1_O,
    WRITEBACK_D_WITH_ALU_T1_O,
    WRITEBACK_D_WITH_PCP1_T1_O,
    WRITEBACK_FLAGS_T1_O,
    WRITEBACK_PC_WITH_PCP1_T1_O,
    WRITEBACK_PC_WITH_S_T1_O,
    
    /* T 2 */
    FETCH_OPCODE_T2_O,
    FETCH_SD_T2_O,
    TFR_SD_T2_O,
    WRITEBACK_D_WITH_ALU_T2_O,
    WRITEBACK_D_WITH_PCP1_T2_O,
    WRITEBACK_FLAGS_T2_O,
    WRITEBACK_PC_WITH_PCP1_T2_O,
    WRITEBACK_PC_WITH_S_T2_O
    );
 
input CLK_I, RESET_I;
input [11:0] OPCODE_T1_I, OPCODE_T2_I;
input FLAG_Z_T1_I,
      FLAG_C_T1_I,
      FLAG_Z_T2_I,
      FLAG_C_T2_I;
    
output reg FETCH_OPCODE_T1_O,
    FETCH_SD_T1_O,
    TFR_SD_T1_O,
    WRITEBACK_D_WITH_ALU_T1_O,
    WRITEBACK_D_WITH_PCP1_T1_O,
    WRITEBACK_FLAGS_T1_O,
    WRITEBACK_PC_WITH_PCP1_T1_O,
    WRITEBACK_PC_WITH_S_T1_O,
    
    /* T 2 */
    FETCH_OPCODE_T2_O,
    FETCH_SD_T2_O,
    TFR_SD_T2_O,
    WRITEBACK_D_WITH_ALU_T2_O,
    WRITEBACK_D_WITH_PCP1_T2_O,
    WRITEBACK_FLAGS_T2_O,
    WRITEBACK_PC_WITH_PCP1_T2_O,
    WRITEBACK_PC_WITH_S_T2_O;
    
    
reg [3:0] state;
reg t2_active;
`define ST_COLD_RESET  0
`define ST_COLD_RESET1 1
`define ST_WARM_RESET  2
`define ST_T1_FETCH    3 /* only T1 fetches */
`define ST_T1_FETCH_SD 4 /* T1 fetches S&D from memory, saves opcode into reg */
`define ST_T2_WB       5
`define ST_T2_FETCH_SD 6
`define ST_T1_WB       7
`define T1_SEQ_OOO     OPCODE_T1_I[11:7]
`define T1_SEQ_CCC     OPCODE_T1_I[6:4]
`define T1_SEQ_Z       OPCODE_T1_I[5]
`define T1_SEQ_C       OPCODE_T1_I[4]
`define T1_SEQ_I       OPCODE_T1_I[0]
`define T2_SEQ_OOO     OPCODE_T2_I[11:7]
`define T2_SEQ_CCC     OPCODE_T2_I[6:4]
`define T2_SEQ_Z       OPCODE_T2_I[5]
`define T2_SEQ_C       OPCODE_T2_I[4]
`define T2_SEQ_I       OPCODE_T2_I[0]


wire t1_execute; /* is TRUE when the instruction has to be executed */
wire t2_execute; /* because the flags allow it or they are not tested */

assign t1_execute = `T1_SEQ_CCC == 0 ? 1'b0:
                    `T1_SEQ_CCC < 3'h4 ? FLAG_Z_T1_I == `T1_SEQ_Z:
                    `T1_SEQ_CCC == 3'h7 ? 1'b1:FLAG_C_T1_I == `T1_SEQ_C;

assign t2_execute = `T2_SEQ_CCC == 0 ? 1'b0:
                    `T2_SEQ_CCC < 3'h4 ? FLAG_Z_T2_I == `T2_SEQ_Z:
                    `T2_SEQ_CCC == 3'h7 ? 1'b1:FLAG_C_T2_I == `T2_SEQ_C;

always @(posedge CLK_I)
	begin
		if (RESET_I == 1'b0)
			state <= 3'b000;
		else
			begin
				if (state[2] == 1'b1)
					state[1:0] <= `RDELAY state + 2'b01;
				else
					state <= `RDELAY state + 3'b01;
				if (state[2:0] == 3'b111)
				    t2_active <= 1'b1;
			end
	end

always @(state)
    begin
        #5
        FETCH_OPCODE_T1_O            = 1'b0;
        FETCH_SD_T1_O                = 1'b0;
        TFR_SD_T1_O                  = 1'b0;
        WRITEBACK_D_WITH_ALU_T1_O    = 1'b0;
        WRITEBACK_D_WITH_PCP1_T1_O   = 1'b0;
        WRITEBACK_FLAGS_T1_O         = 1'b0;
        WRITEBACK_PC_WITH_PCP1_T1_O  = 1'b0;
        WRITEBACK_PC_WITH_S_T1_O     = 1'b0;
    
        /* T 2 */
        FETCH_OPCODE_T2_O            = 1'b0;
        FETCH_SD_T2_O                = 1'b0;
        TFR_SD_T2_O                  = 1'b0;
        WRITEBACK_D_WITH_ALU_T2_O    = 1'b0;
        WRITEBACK_D_WITH_PCP1_T2_O   = 1'b0;
        WRITEBACK_FLAGS_T2_O         = 1'b0;
        WRITEBACK_PC_WITH_PCP1_T2_O  = 1'b0;
        WRITEBACK_PC_WITH_S_T2_O     = 1'b0;
        case (state)
            `ST_COLD_RESET: begin end
			`ST_COLD_RESET1: begin end			
			`ST_WARM_RESET: begin end
			`ST_T1_FETCH:
				FETCH_OPCODE_T1_O = 1'b1;
			`ST_T1_FETCH_SD: /* 4 */
				begin
					FETCH_SD_T1_O = 1'b1;
					if (t2_active)
					    begin
					        WRITEBACK_PC_WITH_PCP1_T2_O = 1'b1;
					        case (`T1_SEQ_OOO)
					             `OP_JUMP_JMP: // JMP or CALL
					                 WRITEBACK_PC_WITH_S_T1_O = 1'b1;
			                endcase
			            end
				end
			`ST_T2_WB: /* 5 */
			    begin
			        TFR_SD_T1_O = 1'b1; // save fetched S&D to temporal registers
			        
			        /* T2, write back  */
			        FETCH_OPCODE_T2_O = 1'b1;
			        if (t2_execute && t2_active)
			            case (`T2_SEQ_OOO)
			                `OP_JUMP_JMP:
			                    begin
			                        WRITEBACK_PC_WITH_S_T2_O = 1'b1;
			                        if (`T2_SEQ_I == 1'b0) // not set, it's a call, write PC+1 to [d]
			                            begin
			                                WRITEBACK_D_WITH_PCP1_T2_O = 1'b1;
			                                //WRITEBACK_PC_WITH_S_T2_O = 1'b1;
			                            end
			                    end
							default:
							    begin
							        WRITEBACK_D_WITH_ALU_T2_O = 1'b1;
							        WRITEBACK_FLAGS_T2_O = 1'b1;
							    end
			            endcase
			                
			    end
			`ST_T2_FETCH_SD:
			    begin
			        FETCH_SD_T2_O = 1'b1;
			        // process jumps, in the next cycle the opcode will be fetched
			        WRITEBACK_PC_WITH_PCP1_T1_O = 1'b1;
			        case (`T1_SEQ_OOO)
			            `OP_JUMP_JMP: // JMP or CALL
			                begin
			                    WRITEBACK_PC_WITH_S_T1_O = 1'b1;
			                end
			        endcase
			        
			    end
			`ST_T1_WB:
			    begin
			        /* T1, write back  */
			        FETCH_OPCODE_T1_O = 1'b1;
			        
			        if (t1_execute)
			            case (`T1_SEQ_OOO)
			                `OP_JUMP_JMP:
			                    begin
			                        if (`T1_SEQ_I == 1'b0) // not set, it's a call, write PC+1 to [d]
			                            begin
			                                WRITEBACK_D_WITH_PCP1_T1_O = 1'b1;
			                            end
			                    end
							default:
							    begin
							        WRITEBACK_D_WITH_ALU_T1_O = 1'b1;
							        WRITEBACK_FLAGS_T1_O = 1'b1;
							    end
			            endcase
			        /* T2 load tempral registers with S&D */
			        TFR_SD_T2_O = 1'b1; // save fetched S&D to temporal registers
			    end
		endcase
	end

initial
    state = 0;
	
endmodule
			
module pnut_v8_mem(
	CLK_I,
	RESET_I,
	
	FETCH_OP_T1_I,
	FETCH_OP_T2_I,
	LOAD_OP_T1_I,
	LOAD_OP_T2_I,
	FETCH_SD_I,
	TFR_SD_I,
	WB_D_T1_I,
	WB_D_T2_I,
	
	D_I,
	PC_T1_I,
	PC_T2_I,
	OPCODE_T1_O,
	OPCODE_T2_O,
	PREOPCODE_T1_O,
	PREOPCODE_T2_O,
	
	D_O,
	S_O
	);

parameter DWIDTH = 32;
parameter MEMBITS = 10;

input CLK_I, /* main clock, rising edge used to load registers */
	RESET_I,   
	FETCH_OP_T1_I, /* fetch from memory opcode */
	FETCH_OP_T2_I,
	LOAD_OP_T1_I, /* signals when the opcode register has to be loaded with the value from memory */
	LOAD_OP_T2_I,/* signals when the opcode register has to be loaded with the value from memory */
	FETCH_SD_I,
	TFR_SD_I,
	WB_D_T1_I,
	WB_D_T2_I;
	
input [31:0] D_I;
input [MEMBITS-1:0] PC_T1_I, PC_T2_I;
output [31:0] S_O, D_O;
output [31:0] OPCODE_T1_O, OPCODE_T2_O, PREOPCODE_T1_O, PREOPCODE_T2_O;

/* Addresses for d and s are taken directly from the opcode */
wire [MEMBITS-1:0] port_a_addr, port_b_addr;
wire [DWIDTH-1:0] port_a_bus, port_b_bus;

wire WB_D;

reg	[DWIDTH-1:0] opcode_t1_r, opcode_t2_r;
reg [DWIDTH-1:0] d_r, s_r;

assign D_O = d_r; /* one for both threads */
assign S_O = s_r;
assign OPCODE_T1_O = opcode_t1_r;
assign OPCODE_T2_O = opcode_t2_r;
assign PREOPCODE_T1_O = port_a_bus;
assign PREOPCODE_T2_O = port_a_bus;

assign WB_D = WB_D_T1_I | WB_D_T2_I;

`ifdef LATTICE
iram_1k_32 imem(.DataInA(), .DataInB(d_in), .AddressA(port_a_addr), 
                .AddressB(read_sd_i ? `ROP_BIT_D:reg_opcode[19:10]), 
                .ClockA(CLK_I), .ClockB(CLK_I), 
                .ClockEnA(1'b1), .ClockEnB(1'b1), .WrA(1'b0), .WrB(save_mem_i), 
				.ResetA(RESET_I), .ResetB(RESET_I), .QA(port_a_bus), .QB(port_b_bus));

`else
memblock_dp_x32 imem(
	.clk_i(CLK_I),
	.port_a_addr_i(port_a_addr),
	.port_b_addr_i(port_b_addr),
	.port_c_addr_i(port_b_addr),
	.port_a_rden_i(FETCH_OP_T1_I | FETCH_OP_T2_I | FETCH_SD_I),
	.port_b_rden_i(FETCH_SD_I),
	.port_a_q_o(port_a_bus),
	.port_b_q_o(port_b_bus),
	.port_c_wen_i(WB_D),
	.port_c_data_i(D_I)
	);
`endif

assign port_a_addr = FETCH_OP_T1_I == 1'b1 ? PC_T1_I:
                     FETCH_OP_T2_I == 1'b1 ? PC_T2_I:
                     port_a_bus[9:0]; /* field s of opcode */
assign port_b_addr = WB_D_T1_I == 1'b1 ? opcode_t1_r[19:10]:
                     WB_D_T2_I == 1'b1 ? opcode_t2_r[19:10]:
                     port_a_bus[19:10]; /* field d of opcode */

always @(posedge CLK_I)
	begin
		if (LOAD_OP_T1_I == 1'b1)
			opcode_t1_r <= port_a_bus;
		if (LOAD_OP_T2_I == 1'b1)
			opcode_t2_r <= port_a_bus;
		if (TFR_SD_I == 1'b1)
		    begin
		        if (opcode_t1_r[`BIT_I] == 1'b1)
		            s_r <= { 22'h0, opcode_t1_r[`FIELD_S] };
		        else
		            s_r <= port_a_bus;
		        d_r <= port_b_bus;
		    end
		
	end
endmodule


module memblock_dp_x32(
	clk_i,
	port_a_addr_i,
	port_b_addr_i,
	port_c_addr_i,
	port_a_rden_i,
	port_b_rden_i,
	port_a_q_o,
	port_b_q_o,
	port_c_wen_i,
	port_c_data_i
	);

input [9:0] port_a_addr_i, port_b_addr_i, port_c_addr_i;
input clk_i, port_a_rden_i, port_b_rden_i, port_c_wen_i;
input [31:0] port_c_data_i;
output reg [31:0] port_a_q_o, port_b_q_o;

reg [31:0] mem[1023:0];

always @(posedge clk_i)
	begin
		if (port_a_rden_i)
			port_a_q_o <= mem[port_a_addr_i];
		if (port_b_rden_i)
			port_b_q_o <= mem[port_b_addr_i];
		if (port_c_wen_i)
			mem[port_c_addr_i] <= port_c_data_i;
	end
integer i;
initial
	begin
		#0
        for (i = 0; i < 1024; i = i+1)
            mem[i] = 0;
		//               op   CCC WWW      631        src
		//                     CZ ZCR i   84268421   84218421
		mem[   0] = 32'b11000_111_001_0_0000100010_0000000100; // call $22, $04
		mem[   1] = 32'b00010_111_001_1_0001000100_0000000010; // and rt, mask
		mem[   2] = 32'b01110_111_001_1_0001000101_0001000011; // mov rt2, rB2
		mem[   3] = 32'b00010_111_001_1_0001000101_0001000110; // and rt2, mask
		mem[   4] = 32'b11000_111_111_1_0001000100_0000000001; // sub rt, rt2 wc
		mem[   5] = 32'b01010_111_001_0_0001000001_0001000110; // andn rA2, mask
		mem[   6] = 32'b01010_111_001_0_0001000011_0001000110; // andn rB2, mask
		mem[   7] = 32'b11000_011_000_1_0000000000_0000010000; // jz  #add
		mem[   8] = 32'b00101_111_110_1_0001000100_0000010000; // cmp rt, #16 
		mem[   9] = 32'b11000_100_000_1_0000000000_0000010000; // jnc #20
		mem[  10] = 32'b10101_111_011_0_0001000010_0001000010; // shr4 rB, rB
		mem[  11] = 32'b10111_111_001_0_0001000011_0001000011; // ror4 rB2, rB2
		mem[  12] = 32'b00000_000_000_0_0000000000_0000000000;
		mem[  13] = 32'b00000_000_000_0_0000000000_0000000000;
		mem[  14] = 32'b00000_000_000_0_0000000000_0000000000;
		mem[  15] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h10] = 32'b00000_000_001_0_0001010000_0001010000;
		mem['h11] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h12] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h13] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h14] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h15] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h16] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h17] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h18] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h19] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h1a] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h1b] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h1c] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h1d] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h1e] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h1f] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h20] = 32'b11000_000_001_0_0001100000_0001100000;
		mem['h21] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h22] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h23] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h24] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h25] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h26] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h27] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h28] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h29] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h2a] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h2b] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h2c] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h2d] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h2e] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h2f] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h30] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h31] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h32] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h33] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h34] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h35] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h36] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h37] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h38] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h39] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h3a] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h3b] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h3c] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h3d] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h3e] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h3f] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h40] = 32'h09876543; // 32'b00000_000_000_0_0000000000_0000000000;
		mem['h41] = 32'h21098817; // 32'b00000_000_000_0_0000000000_0000000000;
		mem['h42] = 32'h05454545; // 32'b00000_000_000_0_0000000000_0000000000;
		mem['h43] = 32'h03232816; // 32'b00000_000_000_0_0000000000_0000000000;
		mem['h44] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h45] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h46] = 32'h00000fff; // 32'b00000_000_000_0_0000000000_0000000000;
		mem['h47] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h48] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h49] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h4a] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h4b] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h4c] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h4d] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h4e] = 32'b00000_000_000_0_0000000000_0000000000;
		mem['h4f] = 32'b00000_000_000_0_0000000000_0000000000;
		
	end
endmodule
/*
                            TERMS OF USE: MIT License

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/
